@@ -55,12 +55,12 @@ class ReadMarkerHandler:
 
             should_update = True
             # Get event ordering, this also ensures we know about the event
-            event_ordering = await self.store.get_event_ordering(event_id)
+            event_ordering = await self.store.get_event_ordering(event_id, room_id)
 
             if existing_read_marker:
                 try:
                     old_event_ordering = await self.store.get_event_ordering(
-                        existing_read_marker["event_id"]
+                        existing_read_marker["event_id"], room_id
                     )
                 except SynapseError:
                     # Old event no longer exists, assume new is ahead. This may
